/*
 * Jeffrey Knockel <jeff250 at unm dot edu> 9/8/09
 * v1.1 9/9/09: Add missing REG_TEST()
 * v1.2 9/18/09: Poll pipe before using a blocking operation
 * v1.3 9/21/09: Fix WRAP_CALL() macro to use no assembler directives
 * v1.4 10/14/09: Add signal handling to catch SIG_ABRT from assert()
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <poll.h>
#include <signal.h>
#include <setjmp.h>

static struct sigaction abrt_act;
static jmp_buf env;
static volatile int errors;

#define TEST(x, env) \
        do { \
            if (setjmp (env) == 0) { \
                (x); \
            } else { \
                ++errors; \
            } \
        } while (0)

/* Definitions of the functions that you have to implement. */

char *my_strcpy( char *dest, const char *src );
char *my_strcat( char *dest, const char *src );
int my_puts( const char *s );
int my_strcmp( const char *s1, const char *s2 );
size_t my_strlen( const char *s );
void my_itoa( int n, char s[] );
void my_reverse( char s[] );
void my_bubblesort( int *a, int n );
void my_swap( int *a, int base1, int base2 );

/* Tests... */

/*
 * This macro tests to make sure you preserve the registers that you are
 * supposed to, after a call to your function.  This is called from each of the
 * test functions below.
 */
#define REG_TEST(stmt)                                      \
do {                                                        \
    register unsigned int _s0 __asm__ ("s0") = 0xdeadbeef;  \
    register unsigned int _s1 __asm__ ("s1") = 0xdeadbeef;  \
    register unsigned int _s2 __asm__ ("s2") = 0xdeadbeef;  \
    register unsigned int _s3 __asm__ ("s3") = 0xdeadbeef;  \
    register unsigned int _s4 __asm__ ("s4") = 0xdeadbeef;  \
    register unsigned int _s5 __asm__ ("s5") = 0xdeadbeef;  \
    register unsigned int _s6 __asm__ ("s6") = 0xdeadbeef;  \
    register unsigned int _s7 __asm__ ("s7") = 0xdeadbeef;  \
    register void * _sp __asm__ ("sp");                     \
    register void * _fp __asm__ ("fp");                     \
                                                            \
    void * _sp_old = _sp;                                   \
    void * _fp_old = _fp;                                   \
                                                            \
    stmt;                                                   \
    assert (_s0 == 0xdeadbeef);                             \
    assert (_s1 == 0xdeadbeef);                             \
    assert (_s2 == 0xdeadbeef);                             \
    assert (_s3 == 0xdeadbeef);                             \
    assert (_s4 == 0xdeadbeef);                             \
    assert (_s5 == 0xdeadbeef);                             \
    assert (_s6 == 0xdeadbeef);                             \
    assert (_s7 == 0xdeadbeef);                             \
    assert (_sp == _sp_old);                                \
    assert (_fp == _fp_old);                                \
} while (0)

static void test_my_strcpy( void )
{
    char dest[20];
    fprintf (stderr, "Testing my_strcpy()...\n");

    REG_TEST(my_strcpy (dest, "foo"));
    assert (strcmp (dest, "foo") == 0);

    my_strcpy (dest, "barf");
    assert (strcmp (dest, "barf") == 0);

    my_strcpy (dest, "barf");
    assert (strcmp (dest, "barf") == 0);

    my_strcpy (dest, "");
    assert (strcmp (dest, "") == 0);
}

static void test_my_strcat( void )
{
    char dest[20];
    fprintf (stderr, "Testing my_strcat()...\n");

    strcpy (dest, "hello ");
    REG_TEST(my_strcat (dest, "world"));

    strcpy (dest, "hello ");
    my_strcat (dest, "world");
    assert (strcmp (dest, "hello world") == 0);

    strcpy (dest, "");
    my_strcat (dest, "");
    assert (strcmp (dest, "") == 0);

    strcpy (dest, "");
    my_strcat (dest, "foo");
    assert (strcmp (dest, "foo") == 0);

    strcpy (dest, "foo");
    my_strcat (dest, "");
    assert (strcmp (dest, "foo") == 0);
}

static void test_my_strcmp( void )
{
    fprintf (stderr, "Testing my_strcmp()...\n");

    REG_TEST(my_strcmp ("foo", "foo"));

    assert (my_strcmp ("foo", "foo") == 0);

    assert (my_strcmp ("barf", "barf") == 0);

    assert (my_strcmp ("foo", "fo") > 0);

    assert (my_strcmp ("fo", "foo") < 0);

    assert (my_strcmp ("fop", "foo") > 0);

    assert (my_strcmp ("foo", "fop") < 0);

    assert (my_strcmp ("", "") == 0);
}

static void test_my_puts( void )
{
    char buff[20];
    int old_stdout;
    int pipefd[2];
    int i;

    struct pollfd pfd;

    fprintf (stderr, "Testing my_puts()...\n");

    REG_TEST(my_puts ("Testing my_puts() register preservation..."));
    fflush (stdout);

    // redirect stdout to a pipe...
    assert ((old_stdout = dup (STDOUT_FILENO)) != -1);
    assert (pipe (pipefd) == 0);
    assert (dup2 (pipefd[1], STDOUT_FILENO) == STDOUT_FILENO);

    // set up fds
    pfd.fd = pipefd[0];
    pfd.events = POLLIN;

    assert (my_puts ("hello world") >= 0);
    fflush (stdout);
    for (i = 0; i < 20; ++i) {
        if (poll (&pfd, 1, 1) <= 0) {
            break;
        }
        read (pipefd[0], buff + i, 1);
    }
    assert (strncmp ("hello world\n", buff, 12) == 0);
    assert (i == 12);

    assert (my_puts ("") >= 0);
    fflush (stdout);
    for (i = 0; i < 20; ++i) {
        poll (&pfd, 1, 100);
        if (poll (&pfd, 1, 1) <= 0) {
            break;
        }
        read (pipefd[0], buff + i, 1);
    }
    assert (strncmp ("\n", buff, 1) == 0);
    assert (i == 1);

    // restore stdout...
    close (pipefd[0]);
    close (pipefd[1]);
    assert (dup2 (old_stdout, STDOUT_FILENO) == STDOUT_FILENO);
}

static void test_my_strlen( void )
{
    fprintf (stderr, "Testing my_strlen()...\n");

    REG_TEST(my_strlen ("hello world"));

    assert (my_strlen ("foo") == 3);

    assert (my_strlen ("beer") == 4);

    assert (my_strlen ("") == 0);
}

static void test_my_itoa( void )
{
    char buff[20];

    fprintf (stderr, "Testing my_itoa()...\n");

    REG_TEST(my_itoa (20, buff));

    my_itoa (20, buff);
    assert (strcmp (buff, "20") == 0);

    my_itoa (0, buff);
    assert (strcmp (buff, "0") == 0);

    my_itoa (-1, buff);
    assert (strcmp (buff, "-1") == 0);

    my_itoa (123456789, buff);
    assert (strcmp (buff, "123456789") == 0);

    my_itoa (-123456789, buff);
    assert (strcmp (buff, "-123456789") == 0);
}

static void test_my_reverse( void )
{
    char buff[20];

    fprintf (stderr, "Testing my_reverse()...\n");

    strcpy (buff, "hello");
    REG_TEST(my_reverse (buff));

    strcpy (buff, "hello");
    my_reverse (buff);
    assert (strcmp (buff, "olleh") == 0);

    strcpy (buff, "tenet");
    my_reverse (buff);
    assert (strcmp (buff, "tenet") == 0);

    strcpy (buff, "");
    my_reverse (buff);
    assert (strcmp (buff, "") == 0);
}

static void test_my_bubblesort( void )
{
    int ary[10] = {4, 5, 0, 3, 9, 8, 2, 7, 1, 6};
    int ary2[100] = {54, 18, 58, 31, 82, 86, 32, 10, 63, 34, 2, 5, 75, 71, 96, 76, 57, 65, 97, 17, 92, 6, 28, 14, 44, 44, 90, 27, 52, 22, 7, 2, 67, 47, 55, 9, 95, 36, 89, 13, 83, 26, 7, 48, 49, 44, 8, 21, 3, 27, 82, 67, 46, 67, 3, 3, 76, 92, 58, 89, 6, 76, 2, 56, 96, 63, 70, 18, 80, 33, 42, 66, 9, 92, 52, 97, 56, 13, 75, 75, 53, 44, 53, 5, 7, 15, 60, 22, 70, 50, 81, 26, 32, 31, 4, 51, 29, 28, 10, 78};
    int i;

    fprintf (stderr, "Testing my_bubblesort()...\n");
    
    REG_TEST(my_bubblesort (ary, 10));

    for (i = 0; i < 10; ++i) {
        assert (ary[i] == i);
    }

    my_bubblesort (ary, 0);

    for (i = 0; i < 10; ++i) {
        assert (ary[i] == i);
    }

    my_bubblesort (ary2, 100);

    for (i = 0; i < 99; ++i) {
        assert (ary2[i] <= ary2[i + 1]);
    }
}

static void test_my_swap( void )
{
    int ary[10] = {0, 1, 2, 3, 4, 5, 6, 7, 8, 9};

    fprintf (stderr, "Testing my_swap()...\n");

    REG_TEST(my_swap (ary, 1, 2));

    my_swap (ary, 0, 8);
    assert (ary[8] == 0);
    assert (ary[0] == 8);

    my_swap (ary, 7, 6);
    assert (ary[7] == 6);
    assert (ary[6] == 7);
    
    my_swap (ary, 4, 4);
    assert (ary[4] == 4);
}

static void abrt_handler( int signum )
{
    (void) signum;
    fprintf (stderr, "Caught abort signal...\n");
    sigaction (SIGABRT, &abrt_act, NULL);
    longjmp (env, 1);
}

int main( void )
{
    abrt_act.sa_handler = &abrt_handler;
    abrt_act.sa_flags = SA_NODEFER;
    sigemptyset (&abrt_act.sa_mask);
    sigaction (SIGABRT, &abrt_act, NULL);

    TEST(test_my_strcpy (), env);
    TEST(test_my_strcat (), env);
    TEST(test_my_puts (), env);
    TEST(test_my_strcmp (), env);
    TEST(test_my_strlen (), env);
    TEST(test_my_itoa (), env);
    TEST(test_my_reverse (), env);
    TEST(test_my_swap (), env);
    TEST(test_my_bubblesort (), env);

    if (errors == 0) {
        fprintf (stderr, "You passed all of the tests!\n");
    } else {
        fprintf (stderr, "You have failed %d tests.\n", errors);
    }

    return 0;
}
